/*****************************************************************************
 *
 * FILE:	mod_session.c
 * DESCRIPTION:	Wyvern: session management module
 * DATE:	Wed, May  5 2004
 * UPDATED:	Mon, Jun  7 2004
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2004 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2004 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_session.c,v 1.2 2004/07/10 06:12:05 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#if	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif	/* HAVE_SYS_TIME_H */
#include <db.h>
#include <wyvern.h>
#include "session.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	SESSION_CONF	DEF_CONF_DIR "/session.conf"
#define	DEF_SDB_DIR	"/tmp"
#define	DEF_SDB_PREFIX	".cc_"
#define	DEF_SDB_SUFFIX	"db"
#define	SESSION_DB \
	DEF_SDB_DIR "/" DEF_SDB_PREFIX SESSION_NAME "." DEF_SDB_SUFFIX

#define WM_SID_INVALID	(-1)
#define WM_SID_VALID	(0)
#define WM_SID_EXPIRED	(1)

typedef struct _wmSession {
  const char *		name;	/* server name */
  const char *		path;	/* region for session use */
  size_t		plen;	/* path length */
  const char *		uri;	/* URI for login (top) page */
  struct _wmSession *	next;
} wmSession;

/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void		wm_session_init(void);
static void		wm_session_finish(void);
static void		wm_session_register_hooks(void);
static int		wm_validate_session(Request *);
static wmSession *	lookup(const char *, const char *);
static int		validate_session_id(const char *);
static int		expire_sid(const DBT *, const DBT *);
static char *		get_sid_from_query(char *);
static char *		get_sid_from_cookie(char *);
static int		parse_string(char *, char **, char **);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static wmSession *	smgr = NULL;
static DB *		sdb  = NULL;

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	wm_session_init
 * Description:		initialize CGI object
 * Arguments:		none
 * Return Value:	none
 */
static void
wm_session_init(void)
{
  register FILE *	fin;

  fin = fopen(SESSION_CONF, "r");
  if (fin) {
    char	buf[BUFSIZ];

#define	MAX_ARGS	(3)
#define	ARGS_SERVER	(0)
#define	ARGS_PATH	(1)
#define	ARGS_URI	(2)
    memset(buf, 0, sizeof(buf));
    while (fgets(buf, sizeof(buf), fin) != NULL) {
      int		n;
      char **		ap;
      char *		args[MAX_ARGS];
      char *		p;
      wmSession *	new;

      if (buf[0] == '#' || buf[0] == SPC || buf[0] == TAB || buf[0] == LF) {
	continue;
      }

      for (n = 0, p = buf, ap = args; (*ap = strsep(&p, " \t\r\n")) != NULL; ) {
	if (**ap != EOL) {
	  n++;
	  if (++ap >= &args[MAX_ARGS]) {
	    break;
	  }
	}
      }
      if (n != MAX_ARGS) {
	continue;
      }

      new = (wmSession *)ws_malloc(sizeof(wmSession));
      if (new) {
	new->name = ws_strdup(args[ARGS_SERVER]);
	new->path = ws_strdup(args[ARGS_PATH]);
	new->plen = strlen(args[ARGS_PATH]);
	new->uri  = ws_strdup(args[ARGS_URI]);
	if (smgr) {
	  new->next = smgr;
	}
	else {
	  new->next = NULL;
	}
	smgr = new;
      }
      memset(buf, 0, sizeof(buf));
    }
    fclose(fin);
  }
}

/*
 * Function Name:	wm_session_finish
 * Description:		delete CGI object
 * Arguments:		none
 * Return Value:	none
 */
static void
wm_session_finish(void)
{
  register wmSession *	sp;
  register wmSession *	sp_next;

  if (sdb) {
    sdb->close(sdb);
  }

  for (sp = smgr; sp != NULL; sp = sp_next) {
    sp_next = sp->next;
    if (sp->name) { free((char *)sp->name); }
    if (sp->path) { free((char *)sp->path); }
    if (sp->uri)  { free((char *)sp->uri); }
    free(sp);
  }
}

/*
 * Function Name:	wm_validate_session
 * Description:		validate a session
 * Arguments:		message
 * Return Value:	Successful: 0, otherwise: -1
 */
static int
wm_validate_session(req)
	register Request *	req;
{
  wmSession *	sp;
  int		ret = -1;

  sp = lookup(req->vhost->serverName, req->uri);
  if (!sp) {
    return 0;
  }

  /* session id in query-string */
  if (req->query_string) {
    char *	s = ws_strdup(req->query_string);
    if (s) {
      char *	id;

      id = get_sid_from_query(s);
      if (id) {
	int	n;

	free((char *)s);
	n = validate_session_id(id);
	if (n == WM_SID_VALID) {
	  return 0;
	}
      }
    }
  }

  /* session id in cookie */
  if (req->header_cookie) {
    char *	s = ws_strdup(req->header_cookie);
    if (s) {
      char *	id;

      id = get_sid_from_cookie(s);
      if (id) {
	int	n;
	char *	c;	/* new cookie */

	free((char *)s);
	n = validate_session_id(id);
#if	DEBUG
	switch (n) {
	  case WM_SID_VALID:
	    wv_syslog(WLL_LOG_DEBUG, "Valid: %s", id);
	    break;
	  case WM_SID_EXPIRED:
	    wv_syslog(WLL_LOG_DEBUG, "Expired: %s", id);
	    break;
	  case WM_SID_INVALID:
	    wv_syslog(WLL_LOG_DEBUG, "Invalid: %s", id);
	    break;
	}
#endif	/* DEBUG */
	if (n == WM_SID_VALID) {
	  asprintf(&c, SESSION_ID "=%s; Path=%s; Version=1", id, sp->path);
	  ret = 0;
	}
	else {
	  asprintf(&c, SESSION_ID "=%s; Max-Age=0; Version=1", id);
	}
	req->header_set_cookie = c;
	free((char *)id);
#if	DEBUG
	wv_syslog(WLL_LOG_DEBUG, "Set-Cookie: %s", c);
#endif	/* DEBUG */
      }
    }
  }

#if	DEBUG
  else {
    wv_syslog(WLL_LOG_DEBUG, "No-Cookie: ");
  }
  wv_syslog(WLL_LOG_DEBUG, "Request: %s:%s", req->vhost->serverName, req->uri);
  wv_syslog(WLL_LOG_DEBUG, "Location: %s", sp->uri);
#endif	/* DEBUG */
  /* invalid session id */
  if (ret == -1) {
    req->header_location = ws_strdup(sp->uri);
    req->status = STATUS_MOVED_PERMANENTLY;
  }

  return ret;
}

static wmSession *
lookup(name, path)
	const char *	name;
	const char *	path;
{
  char *	uri;

  asprintf(&uri, "http://%s%s", name, path);
  if (uri) {
    register wmSession *	sp;

    for (sp = smgr; sp != NULL; sp = sp->next) {
      if (ws_strcmp(uri, sp->uri) != 0 &&
	  ws_strcmp(name, sp->name) == 0 &&
	  ws_strncmp(path, sp->path, sp->plen) == 0) {
	free((char *)uri);
	return sp;
      }
    }
    free((char *)uri);
  }

  return NULL;
}

/*
 * Function Name:	validate_session_id
 * Description:		validate a session id
 * Arguments:		session id
 * Return Value:	Successful: 0, otherwise: -1
 */
static int
validate_session_id(id)
	const char *	id;
{
  if (sdb) {
    sdb->close(sdb);
  }
  sdb = dbopen(SESSION_DB, O_RDWR|O_EXLOCK, 0600, DB_HASH, NULL);
  if (sdb) {
    DBT	key;
    DBT	val;

    key.data = (char *)id;
    key.size = strlen(id);
    if (sdb->get(sdb, &key, &val, 0) == 0) {
      int	ret;

      ret = expire_sid(&key, &val);
      sdb->close(sdb);	sdb = NULL;

      return ret;
    }
#if	DEBUG
    wv_syslog(WLL_LOG_DEBUG, "No exist: %s", id);
#endif	/* DEBUG */
    sdb->close(sdb);	sdb = NULL;
  }

  return WM_SID_INVALID;
}

/*
 * FUNCTION NAME:	expire_sid
 * DESCRIPTION:		expire the session id
 * ARGUMENTS:		key, val
 * RETURN VALUE:	Expired: 1, error: -1, otherwise: 0
 */
static int
expire_sid(key, val)
	const DBT *	key;	/* id */
	const DBT *	val;	/* expired */
{
  struct timeval	now;

  if (gettimeofday(&now, NULL) != -1) {
    char *	s;

    s = (char *)ws_calloc(val->size + 1, sizeof(char));
    if (s) {
      long	timelimit;	/* expired timestamp */

      memcpy(s, val->data, val->size);
      timelimit = strtol(s, (char **)NULL, 10);
      free((char *)s);

      if (now.tv_sec > timelimit) {
	return sdb->del(sdb, key, 0) == 0 ? WM_SID_EXPIRED : WM_SID_INVALID;
      }

      return WM_SID_VALID;
    }
  }

  return WM_SID_INVALID;
}


/*
 * FUNCTION NAME:	get_sid_from_query
 * DESCRIPTION:		parse CGI query data
 * ARGUMENTS:		query string
 * RETURN VALUE:	Successful: id, otherwise: NULL
 */
static char *
get_sid_from_query(query_string)
	char *	query_string;
{
  char *	string = query_string;

  while ((string = strtok(string, "&")) != NULL) {
    char *	name;
    char *	value;

    name  = string;
    value = (char *)strchr(string, '=');
    if (value != NULL) {
      *value++ = EOL;
      if (ws_strcmp(name, SESSION_ID) == 0) {
       	return value;
      }
    }
    string = NULL;	/* DON'T REMOVE THIS LINE. */
  }

  return NULL;
}

/*
 * FUNCTION NAME:	get_sid_from_cookie
 * DESCRIPTION:		parse Cookie string
 * ARGUMENTS:		cookie string
 * RETURN VALUE:	Successful: id, otherwise: NULL
 */
static char *
get_sid_from_cookie(http_cookie)
	char *	http_cookie;
{
  char *	string = http_cookie;

  while ((string = strtok(string, ";,")) != NULL) {
    char *	name;	/* parameter name */
    char *	value;	/* parameter value */

    /* skip white space */
    while (*string == SPC || *string == TAB) {
      string++;
    }

    if (parse_string(string, &name, &value) == 0) {
      if (ws_strcmp(name, SESSION_ID) == 0) {
	free((char *)name);
	return value;
      }
    }
    string = NULL;	/* DON'T REMOVE! */
  }

  return NULL;
}

/*
 * FUNCTION NAME:	parse_string
 * DESCRIPTION:		parse NAME=VALUE string of cookie
 * ARGUMENTS:		string, name pointer, value pointer
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
parse_string(string, name, value)
	register char *	string;
	char **		name;
	char **		value;
{
  register char *	p;

#if	DEBUG
  wv_syslog(WLL_LOG_DEBUG, "Cookie: %s", string);
#endif	/* DEBUG */
  /* initialize */
  *name = *value = NULL;

  /* $Secure and $Discard is special parameter without a value */
  if (ws_strcasecmp(string, "$Secure")  == 0 ||
      ws_strcasecmp(string, "$Discard") == 0) {	/* see RFC2965 */
    *name = ws_strdup(string);
    return 0;
  }

  /* NAME */
  p = strchr(string, '=');
  if (p == NULL) {
    return -1;
  }
  *p++  = EOL;
  *name = ws_strdup(string);
  /* VALUE */
  if (*p == '"') {
    char *	q;

    q = ++p;
    p = strchr(q, '"');
    if (p) {
      *p = EOL;
    }
    *value = ws_strdup(q);
  }
  else {
    *value = ws_strdup(p);
  }

  if (*name == NULL || *value == NULL) {
    wv_free((char *)*name);
    wv_free((char *)*value);
    return -1;
  }

  return 0;
}

/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/
static void
wm_session_register_hooks(void)
{
  wm_hook_fixups(wm_validate_session, HOOK_FIXUP_REQUEST);
}

static wmModule	mod_session = {
  WM_MAGIC_NUMBER_MAJOR,	/* version */
  WM_UNSECURE,			/* unsecure */
  "mod_session",		/* name */
  "Session Management Module version 1.0.1",	/* info */
  "Kouichi ABE (WALL)",		/* author */
  NULL,				/* depend */
  NULL,				/* handle */
  (void *)wm_session_init,		/* init */
  (void *)wm_session_finish,	/* finish */
  (void *)wm_session_register_hooks	/* register_handler */
};

wmModule *
wmModuleInfo(void)
{
  return &mod_session;
}
