/*****************************************************************************
 *
 * FILE:	mod_check.c
 * DESCRIPTION:	Cockatrice: check string utilities (module)
 * DATE:	Thu, Jun 20 2002
 * UPDATED:	Sat, Sep 22 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2002-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2002-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_check.c,v 1.2 2007/09/21 20:15:38 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <regex.h>
#include "cockatrice.h"
#include "module.h"
#include "functions.h"

#if	DEBUG_CHECK
#define	cc_strdup	strdup
#endif	/* DEBUG_CHECK */

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	cm_check_init(void);
static void	cm_check_finish(void);
static void	check_error(regex_t * reg, const char * pat, int errcode);
static int	validate_string(regex_t * reg, const char * s);
static int	cm_check_empty(const char *);
static int	cm_check_digit(const char *);
static int	cm_check_alphabet(const char *);
static int	cm_check_username(const char *);
static int	cm_check_hostname(const char *);
static int	cm_check_domainname(const char *);
static int	cm_check_email(const char *);
static int	cm_check_range(const char *, int, int);
static int	cm_check_ipaddr(const char *);
static int	cm_check_netaddr(const char *);
static int	cm_check_ip_port(const char *);
static int	cm_check_etheraddr(const char *);

/*****************************************************************************
 *
 *	Local variables definition
 *
 *****************************************************************************/
static const char	pat_alphabet[] = "^[a-zA-Z]+$";
static const char	pat_username[] = "^[a-zA-Z][a-zA-Z0-9]+$";
static const char	pat_hostname[] = "^[a-z][-a-z0-9]+[a-z0-9]$";
static const char	pat_domainname[] = "^[a-z][-a-z0-9\\.]+\\.[a-z]{2,4}$";
static const char	pat_email[] = "^[A-Z0-9._%+-]+@[a-z0-9][-A-Z0-9\\.]+\\.[A-Z]{2,4}$";
static const char	pat_ipaddr[] = "^((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$";
static const char	pat_etheraddr[] = "^[0-9a-f]{2}[-:]([0-9a-f]{2}[-:]){4}[0-9a-f]{2}$";

static regex_t		reg_alphabet;
static regex_t		reg_username;
static regex_t		reg_hostname;
static regex_t		reg_domainname;
static regex_t		reg_email;
static regex_t		reg_ipaddr;
static regex_t		reg_etheraddr;

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/
static void
cm_check_init(void)
{
  int	ec;

  ec = regcomp(&reg_alphabet, pat_alphabet, REG_EXTENDED|REG_NOSUB);
  check_error(&reg_alphabet, pat_alphabet, ec);
  ec = regcomp(&reg_username, pat_username, REG_EXTENDED|REG_NOSUB);
  check_error(&reg_username, pat_username, ec);
  ec = regcomp(&reg_hostname, pat_hostname, REG_EXTENDED|REG_ICASE|REG_NOSUB);
  check_error(&reg_hostname, pat_hostname, ec);
  ec = regcomp(&reg_domainname, pat_domainname, REG_EXTENDED|REG_ICASE|REG_NOSUB);
  check_error(&reg_domainname, pat_domainname, ec);
  ec = regcomp(&reg_email, pat_email, REG_EXTENDED|REG_ICASE|REG_NOSUB);
  check_error(&reg_email, pat_email, ec);
  ec = regcomp(&reg_ipaddr, pat_ipaddr, REG_EXTENDED|REG_NOSUB);
  check_error(&reg_ipaddr, pat_ipaddr, ec);
  ec = regcomp(&reg_etheraddr, pat_etheraddr, REG_EXTENDED|REG_ICASE|REG_NOSUB);
  check_error(&reg_etheraddr, pat_etheraddr, ec);
}

static void
cm_check_finish(void)
{
  regfree(&reg_alphabet);
  regfree(&reg_username);
  regfree(&reg_hostname);
  regfree(&reg_domainname);
  regfree(&reg_email);
  regfree(&reg_ipaddr);
  regfree(&reg_etheraddr);
}

static void
check_error(regex_t * reg, const char * pat, int errcode)
{
  if (errcode != 0) {
    char	errbuf[BUFSIZ];
    size_t	size;

    size = regerror(errcode, reg, errbuf, sizeof(errbuf));
    fprintf(stderr, "error: %s %s\n", pat, errbuf);
    exit(1);
  }
}

/*****************************************************************************/

static int
validate_string(reg, s)
	regex_t *	reg;
	const char *	s;	/* test string */
{
  if (s != NULL && s[0] != EOL) {
    int	status;

    status = regexec(reg, s, 0, NULL, 0);

    if (status == 0) {
      return 0;	/* valid */
    }
  }
  return -1;	/* invalid */
}

/*
 * Function Name:	cm_check_empty
 * Description:		check string with empty
 * Arguments:		string
 * Return Value:	Empty: 0, otherwise: -1
 */
static int
cm_check_empty(s)
	const char *	s;
{
  return (s == NULL || s[0] == EOL) ? 0 : -1;
}

/*
 * Function Name:	cm_check_digit
 * Description:		check string with digit
 * Arguments:		string
 * Return Value:	Digit: 0, otherwise: -1
 */
static int
cm_check_digit(s)
	const char *	s;
{
  if (s != NULL && s[0] != EOL) {
    static const char	digits[] = "0123455689";
    size_t		len;
    size_t		pos;

    len = strlen(s);
    pos = strspn(s, digits);
    if (pos == len) {
      return 0;
    }
  }
  return -1;
}

/*
 * Function Name:	cm_check_alphabet
 * Description:		check string with alphabet
 * Arguments:		string
 * Return Value:	Alphabet: 0, otherwise: -1
 */
static int
cm_check_alphabet(s)
	const char *	s;
{
  return validate_string(&reg_alphabet, s);
}

/*
 * Function Name:	cm_check_username
 * Description:		check string with username
 * Arguments:		string
 * Return Value:	UserName: 0, otherwise: -1
 * NOTE:		'^[a-z][a-z0-9]+$'
 */
static int
cm_check_username(s)
	const char *	s;
{
  return validate_string(&reg_username, s);
}

/*
 * Function Name:	cm_check_hostname
 * Description:		check string with hostname
 * Arguments:		string
 * Return Value:	HostName: 0, otherwise: -1
 * NOTE:		'^[a-zA-Z][-a-zA-Z0-9]+[a-zA-Z0-9]$'
 */
static int
cm_check_hostname(s)
	const char *	s;
{
  return validate_string(&reg_hostname, s);
}

/*
 * Function Name:	cm_check_domainname
 * Description:		check string with DomainName
 * Arguments:		string
 * Return Value:	DomainName: 0, otherwise: -1
 */
static int
cm_check_domainname(s)
	const char *	s;
{
  return validate_string(&reg_domainname, s);
}

/*
 * Function Name:	cm_check_email
 * Description:		check string with E-Mail
 * Arguments:		string
 * Return Value:	E-Mail: 0, otherwise: -1
 */
static int
cm_check_email(s)
	const char *	s;
{
  return validate_string(&reg_email, s);
}

/*
 * Function Name:	cm_check_range
 * Description:		check range of string with digit
 * Arguments:		string, min, max
 * Return Value:	successful: 0, otherwise: -1
 */
static int
cm_check_range(s, min, max)
	const char *	s;
	int		min;
	int		max;
{
  if (cm_check_digit(s) == 0 && min < max) {
    int	val;

    val = (int)strtol(s, (char **)NULL, 10);

    return (val >= min && val <= max) ? 0 : -1;
  }
  return -1;
}

/*
 * Function Name:	cm_check_ipaddr
 * Description:		check IP address
 * Arguments:		string
 * Return Value:	successful: 0, otherwise: -1
 */
static int
cm_check_ipaddr(s)
	const char *	s;
{
  return validate_string(&reg_ipaddr, s);
}

/*
 * Function Name:	cm_check_netaddr
 * Description:		check network address
 * Arguments:		string
 * Return Value:	successful: 0, otherwise: -1
 */
static int
cm_check_netaddr(s)
	const char *	s;
{
  int	status = -1;

  if (s != NULL && s[0] != EOL) {
    char *	netaddr;

    netaddr = cc_strdup(s);
    if (netaddr != NULL) {
      char *	masklen;

      masklen = (char *)strrchr(netaddr, '/');
      if (masklen != NULL) {
	*masklen++ = EOL;
	if (cm_check_ipaddr(netaddr) == 0 &&
	    cm_check_range(masklen, 0, 32) == 0) {
  	  status = 0;
   	}
      }
      free(netaddr);
    }
  }

  return status;
}

/*
 * Function Name:	cm_check_ip_port
 * Description:		check IP address and port pair
 * Arguments:		string
 * Return Value:	successful: 0, otherwise: -1
 */
static int
cm_check_ip_port(s)
	const char *	s;
{
  int	status = -1;

  if (s != NULL && s[0] != EOL) {
    char *	ipaddr;

    ipaddr = cc_strdup(s);
    if (ipaddr != NULL) {
      char *	port;

      port = (char *)strrchr(ipaddr, ':');
      if (port != NULL) {
	*port++ = EOL;

	if (cm_check_ipaddr(ipaddr) == 0 &&
	    cm_check_range(port, 0, 65535) == 0) {
	  status = 0;
	}
      }
      free(ipaddr);
    }
  }

  return status;
}

/*
 * Function Name:	cm_check_etheraddr
 * Description:		check ethernet address
 * Arguments:		string
 * Return Value:	successful: 0, otherwise: -1
 */
static int
cm_check_etheraddr(s)
	const char *	s;
{
  return validate_string(&reg_etheraddr, s);
}

/*****************************************************************************/

#ifndef	DEBUG_CHECK
static const cmCheck	check_funcs = {
  cm_check_empty,
  cm_check_digit,
  cm_check_alphabet,
  cm_check_username,
  cm_check_hostname,
  cm_check_domainname,
  cm_check_email,
  cm_check_range,
  cm_check_ipaddr,
  cm_check_netaddr,
  cm_check_ip_port,
  cm_check_etheraddr
};

static void
cm_check_register_hooks(void)
{
  cc_hook_module_check(&check_funcs);
}

static ccModule	mod_check = {
  CC_MODULE_MAGIC_NUMBER_MAJOR,		/* API version */
  "mod_check",				/* module name */
  "Cockatrice: check module 1.3",	/* description such as version etc.. */
  NULL,					/* module handle */
  (void *)cm_check_init,		/* called when it's loaded */
  (void *)cm_check_finish,		/* called when it's unloaded */
  (void *)cm_check_register_hooks
};

ccModule *
ccModuleInfo(void)
{
  return &mod_check;
}
#endif	/* DEBUG_CHECK */

/*****************************************************************************/

#if	DEBUG_CHECK
int
main(argc, argv)
	int	argc;
	char *	argv[];
{
  if (argc == 1) {
    fprintf(stderr, "usage: <string>\n");
    exit(1);
  }
  cm_check_init();
  if (cm_check_empty(*++argv) == 0) {
    fprintf(stderr, "%s: empty\n", *argv);
  }
  if (cm_check_digit(*argv) == 0) {
    fprintf(stderr, "%s: digit\n", *argv);
  }
  if (cm_check_alphabet(*argv) == 0) {
    fprintf(stderr, "%s: alphabet\n", *argv);
  }
  if (cm_check_username(*argv) == 0) {
    fprintf(stderr, "%s: username\n", *argv);
  }
  if (cm_check_hostname(*argv) == 0) {
    fprintf(stderr, "%s: hostname\n", *argv);
  }
  if (cm_check_domainname(*argv) == 0) {
    fprintf(stderr, "%s: domainname\n", *argv);
  }
  if (cm_check_email(*argv) == 0) {
    fprintf(stderr, "%s: e-mail\n", *argv);
  }
  if (cm_check_range(*argv, 0, 255) == 0) {
    fprintf(stderr, "%s: range\n", *argv);
  }
  if (cm_check_ipaddr(*argv) == 0) {
    fprintf(stderr, "%s: ipaddr\n", *argv);
  }
  if (cm_check_netaddr(*argv) == 0) {
    fprintf(stderr, "%s: netaddr\n", *argv);
  }
  if (cm_check_ip_port(*argv) == 0) {
    fprintf(stderr, "%s: ipaddr+port\n", *argv);
  }
  if (cm_check_etheraddr(*argv) == 0) {
    fprintf(stderr, "%s: etheraddr\n", *argv);
  }
  else {
    fprintf(stderr, "%s: unknown\n", *argv);
  }
  cm_check_finish();

  return 0;
}
#endif	/* DEBUG_CHECK */
